--
--  Time tagger logic.
--
--  The time tagger assigns timestamps to events on the digital inputs.
--  It produces a stream of event records to the DMA write channel.
--
--  Output record format:
--    bits 63 : 60  = record type
--    bits 59 : 56  = channel index (only for event record)
--    bits 55 : 52  = 0000
--    bits 51 : 48  = digital signal state (or 0 for overflow record)
--    bits 47 : 0   = timestamp (or 0 for overflow record)
--
--  Channel index:
--    Even channel index (2*k) represents a rising edge on channel k.
--    Odd channel index (2*k+1) represents a falling edge on channel k.
--
--  Joris van Rantwijk 2024
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

use work.puzzlefw_pkg.all;


entity timetagger is

    port (
        -- Main clock, active on rising edge.
        clk:                in  std_logic;

        -- Reset, active high, synchronous to main clock.
        reset:              in  std_logic;

        -- Channel enable mask.
        channel_en:         in  std_logic_vector(7 downto 0);

        -- High to emit a marker record.
        -- An arbitrary delay may pass between a marker request and actual
        -- emitting of the marker record.
        marker:             in  std_logic;

        -- Global timestamp counter.
        timestamp_in:       in  std_logic_vector(timestamp_bits - 1 downto 0);

        -- Digital input signals.
        dig_sample:         in  std_logic_vector(3 downto 0);

        -- Output data stream.
        out_valid:          out std_logic;
        out_ready:          in  std_logic;
        out_empty:          in  std_logic;
        out_data:           out dma_data_type
    );

end entity;

architecture arch of timetagger is

    type regs_type is record
        overflow:           std_logic;
        marker_pending:     std_logic;
        marker_holdoff:     std_logic;
        cur_sample:         std_logic_vector(3 downto 0);
        prev_sample:        std_logic_vector(3 downto 0);
        detected_events:    std_logic_vector(3 downto 0);
        pending_events:     std_logic_vector(3 downto 0);
        fifo_in_valid:      std_logic;
        fifo_in_evmask:     std_logic_vector(3 downto 0);
        out_valid:          std_logic;
        out_data:           dma_data_type;
    end record;

    constant regs_init: regs_type := (
        overflow            => '0',
        marker_pending      => '0',
        marker_holdoff      => '0',
        cur_sample          => (others => '0'),
        prev_sample         => (others => '0'),
        detected_events     => (others => '0'),
        pending_events      => (others => '0'),
        fifo_in_valid       => '0',
        fifo_in_evmask      => (others => '0'),
        out_valid           => '0',
        out_data            => (others => '0')
    );

    signal r: regs_type := regs_init;
    signal rnext: regs_type;

    signal s_fifo_in_ready:     std_logic;
    signal s_fifo_in_data:      std_logic_vector(55 downto 0);
    signal s_fifo_out_valid:    std_logic;
    signal s_fifo_out_ready:    std_logic;
    signal s_fifo_out_data:     std_logic_vector(55 downto 0);

    -- Return the index of the first active channel.
    function find_first_active_channel(ev: std_logic_vector; st: std_logic_vector)
        return std_logic_vector
    is
        constant x: std_logic_vector(ev'length - 1 downto 0) := ev;
        constant y: std_logic_vector(st'length - 1 downto 0) := st;
        variable r: std_logic_vector(3 downto 0) := "0000";
    begin
        for i in 0 to x'high loop
            if x(i) = '1' then
                r(3 downto 1) := std_logic_vector(to_unsigned(i, 3));
                r(0) := not y(i);
                exit;
            end if;
        end loop;
        return r;
    end function;

    -- Clear the least significant non-zero bit.
    function clear_first_nonzero(ev: std_logic_vector)
        return std_logic_vector
    is
        variable r: std_logic_vector(ev'length - 1 downto 0);
    begin
        r := ev;
        for i in 0 to r'high loop
            if r(i) = '1' then
                r(i) := '0';
                exit;
            end if;
        end loop;
        return r;
    end function;

begin

    -- Small FIFO buffer for events.
    -- A single entry in this buffer may expand into multiple event records
    -- thus taking up multiple clock cycles.
    inst_event_fifo: entity work.simple_fifo
        generic map (
            data_width      => 56,
            fifo_depth_bits => 4,
            block_ram       => false )
        port map (
            clk             => clk,
            reset           => reset,
            in_valid        => r.fifo_in_valid,
            in_ready        => s_fifo_in_ready,
            in_data         => s_fifo_in_data,
            out_valid       => s_fifo_out_valid,
            out_ready       => s_fifo_out_ready,
            out_data        => s_fifo_out_data,
            queue_length    => open );

    -- Drive data to FIFO.
    --   bits 55 : 52 = mask of active channels
    --   bits 51 : 48 = new input state after events
    --   bits 47 : 0  = timestamp
    s_fifo_in_data <= r.fifo_in_evmask & r.prev_sample & timestamp_in;

    -- Read from FIFO if we can produce an output record and there
    -- are no pending events.
    s_fifo_out_ready <= out_ready and (not or_reduce(r.pending_events));

    -- Drive output ports.
    out_valid   <= r.out_valid;
    out_data    <= r.out_data;

    --
    -- Combinatorial process.
    --
    process (all) is
        variable v: regs_type;
    begin
        -- Load current register values.
        v := r;

        -- Latch marker request.
        if marker = '1' then
            v.marker_pending := '1';
        end if;

        -- Hold off marker record emission during the cycle following insertion
        -- of an event into the FIFO. This prevents the marker record from
        -- cutting in line and appearing before an event record with an
        -- earlier timestamp.
        v.marker_holdoff := r.fifo_in_valid;

        --
        -- Detect events and write to FIFO.
        --
        -- This logic delays the digital input path in order to match the delay
        -- in the analog acquisition chain. An event timestamp is assigned
        -- after a delay of 3 clock cycles following a change in "dig_sample".
        -- 

        -- Capture input state.
        v.cur_sample := dig_sample;

        -- Keep previous input state.
        v.prev_sample := r.cur_sample;

        -- Detect events on enabled channels.
        for i in 0 to 3 loop
            v.detected_events(i) :=
                (channel_en(2*i) and (not r.prev_sample(i)) and r.cur_sample(i))
                or (channel_en(2*i+1) and r.prev_sample(i) and (not r.cur_sample(i)));
        end loop;

        -- Write detected events to FIFO.
        v.fifo_in_evmask := r.detected_events;
        if (r.overflow = '0') and (or_reduce(r.detected_events) = '1') then
            v.fifo_in_valid := '1';
        else
            v.fifo_in_valid := '0';
        end if;

        -- Detect FIFO overflow.
        if (r.fifo_in_valid = '1') and (s_fifo_in_ready = '0') then
            v.overflow := '1';
            v.fifo_in_valid := '0';
        end if;

        --
        -- Read from FIFO and generate event records.
        --

        if out_ready = '1' then

            -- By default, do not generate a new record.
            v.out_valid := '0';

            if or_reduce(r.pending_events) = '1' then

                -- Emit the next event of a series of simultaneous events.
                v.out_valid := '1';
                v.out_data(59 downto 56) := find_first_active_channel(
                    r.pending_events,
                    r.out_data(51 downto 48));

                v.pending_events := clear_first_nonzero(r.pending_events);

            elsif s_fifo_out_valid = '1' then

                -- Got digital events from FIFO.
                -- Emit the first event. Keep any other events.
                v.out_valid := '1';
                v.out_data(63 downto 60) := msg_timetagger_data;
                v.out_data(59 downto 56) := find_first_active_channel(
                    s_fifo_out_data(55 downto 52),
                    s_fifo_out_data(51 downto 48));
                v.out_data(55 downto 52) := "0000";
                v.out_data(51 downto 48) := s_fifo_out_data(51 downto 48);
                v.out_data(47 downto 0)  := s_fifo_out_data(47 downto 0);

                v.pending_events := clear_first_nonzero(s_fifo_out_data(55 downto 52));

            elsif r.overflow = '1' then

                -- Wait until output buffer empty, then emit overflow record.
                v.out_data(63 downto 56) := msg_overflow;
                v.out_data(55 downto 0)  := (others => '0');
                if out_empty = '1' then
                    v.out_valid := '1';
                    v.overflow  := '0';
                end if;

            elsif (r.marker_pending = '1') and (r.marker_holdoff = '0') then

                -- Emit marker record.
                v.out_valid := '1';
                v.out_data(63 downto 56) := msg_marker;
                v.out_data(55 downto 52) := "0000";
                v.out_data(51 downto 48) := r.prev_sample;
                v.out_data(47 downto 0)  := timestamp_in;

                v.marker_pending := '0';

            end if;

        end if;

        --
        -- Synchronous reset.
        --
        if reset = '1' then
            v := regs_init;
        end if;

        -- Drive new register values to synchronous process.
        rnext <= v;

    end process;

    --
    -- Synchronous process.
    --
    process (clk) is
    begin
        if rising_edge(clk) then
            r <= rnext;
        end if;
    end process;

end architecture;
